# Released under GPL3 terms (see LICENSE)

"""Various input/output functions"""

import os
import sys
import json

def push(output={}, **kwargs):
    """
    Print `output` and flush stdout

    If `output` is dict-like or list-like it is converted to JSON first.
    """
    if output:
        if isinstance(output, (set, tuple)):
            output = list(output)
        if isinstance(output, (dict, list)):
            output = json.dumps(output)
        print(output, flush=True)
    if kwargs:
        print(json.dumps(kwargs), flush=True)


def debug(*msgs):
    """Print `msgs` to stderr"""
    print(' '.join(str(m) for m in msgs), file=sys.stderr, flush=True)


def croak(errmsg, exitcode=1):
    """Print `errmsg` and exit with `exitcode`"""
    print(errmsg, file=sys.stderr, flush=True)
    exit(exitcode)


def get_var(variable, default=None, unset_msg=None, options=()):
    """
    Return value of environment `variable`

    If `default` is None, an error is printed to STDERR if the variable is not
    set. `unset_msg`, if not None, is a custom error message, otherwise one is
    generated.

    If `default` is not None, it will be returned if the variable is not set.

    If the environment variable is set, its value is converted to `default`'s
    type.

    Valid boolean values are true/false, yes/no, on/off and 1/0 (case is
    ignored).

    If `options` is a non-empty iterable, `variable`'s value must be a member
    of it.

    """
    value = os.environ.get(variable, default)
    if value is None:
        if unset_msg is None:
            unset_msg = '{} is not set'.format(variable)
        croak(unset_msg)

    if default is not None:
        # Bools are special because python considers every non-empty string as True.
        if type(default) is bool and options == ():
            if str(value).lower() in ('true', 'yes', 'on', '1'):
                return True
            elif str(value).lower() in ('false', 'no', 'off', '0'):
                return False
            else:
                croak('Invalid boolean value: {!r}'.format(value))
        elif type(value) != type(default):
            try:
                value = type(default)(value)
            except Exception:
                croak('Invalid {} value: {!r}'.format(variable, value))

    if options and value not in options:
        croak('{} must be one of: {}'
              .format(variable, ', '.join(repr(o) for o in options)))

    return value


def read_file(path):
    """Return contents from file `path` (without trailing newline)"""
    try:
        content = open(path, 'r').readline().strip('\n')
    except OSError as e:
        croak('Unable to read {!r}: {}'.format(path, e.strerror))
    return content
